(import :std/error
				:std/sugar
				:std/contract
				:std/foreign
				:std/os/fd
				:std/os/fdio
				:std/os/fcntl
				:std/os/error
				:std/misc/bytes
				:std/srfi/9
				:std/format)

(export #t)

(begin-ffi (EVIOCGRAB KEY_CNT LED_CNT EV_KEY EV_LED _ioctl evt-size)
  (c-declare "#include <linux/input.h>")
	(define-const KEY_CNT)
	(define-const LED_CNT)
	(define-const EV_KEY)
	(define-const EV_LED)
	(define-const EVIOCGRAB)
	(define evt-size ((c-lambda () int "___return(sizeof(struct input_event));")))
	(c-declare "#include <sys/ioctl.h>")
	(define-c-lambda __ioctl (int unsigned-long int) int "ioctl")
	(define-with-errno _ioctl __ioctl (fd request op)))

;; Global state

;; used in log macro, provided by the -v flag
(def verbose #f)

;; raw file descriptor, used by read-ev and led!
(def kb #f)

;; CAUTION - not a boolean, but a C boolean (1 or 0)
(def grabbed? 0)

;; KEY_CNT is 768.
;; On a standard PC keyboard, highest key is Menu, which is 127
;; media buttons can go up as high as 165 OR HIGHER
;; On amd64, this vector can use 6kb of RAM at most
;; there isn't much worth in squeezing memory with this variable, but
;; I included it nonetheless
(def bindable-key-count KEY_CNT)

;; contains the bound functions
(def bindings
	(make-vector bindable-key-count #f))


;; opens file and maybe grabs the keyboard
(def (init-kb filename grab?)
	;; use _open so O_CLOEXEC and O_NONBLOCK are avoided
	(set! kb (_open filename O_RDWR #o644))
	(when grab?
		(thread-sleep! 1)
		;; we can only fail in this if a process has already grabbed
		;; the keyboard. This can happen if WE want to grab it the second
		;; time.
		(grab! 1)))

;; input event abstraction over the linux API
(define-record-type <evt>
	(evt key action)
	evt?
	(key evt-key)
	(action evt-action))


;; this one should be obvious
(def (call-cb ev)
	;; binds the vector-ref if neither is #f
	(awhen (fun (and (< (evt-key ev) bindable-key-count)
									 (vector-ref bindings (evt-key ev))))
		(fun (evt-action ev))))


;; reads a key event from the keyboard
(def (read-ev)
	;; event "struct" used for input
	;; first 16 bytes are time
	;; u16 after that is event type ( SYN, KEY, etc. )
	;; u16 after that is event code ( keycode )
	;; s32 after that is event value ( release, press, repeat )
	(let ((e (make-u8vector evt-size 0)))

		;; we can use fdread because the keyboard ONLY sends data in
		;; `evt-size`-byte packets and if it fails readsing data, we
		;; can be sure that the keyboard has disconnected
		(until (and (fdread kb e)
								;; ignore any non-key event
								(= (u8vector-u16-native-ref e 16) EV_KEY)))
		(evt (u8vector-u16-native-ref e 18)
				 (u8vector-s32-native-ref e 20))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; User-facing functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; build a wrapper around it if you want logging/output
(def (bind! key fun)
	(using ((key :~ (in-range? 0 bindable-key-count))
					(fun :~ procedure?))
		(vector-set! bindings key fun)))


;; check input.h and input-event-codes.h for the led ids
;; op is either 1 (ON) or 0 (OFF)
(def (led! id op)
	(using ((id :~ (in-range? 0 LED_CNT))
					(op :~ (in-range? 0 2)))
		(fdwrite kb (u8vector 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
													EV_LED 0 id 0 op 0 0 0))))


;; use sleep before grabbing, because keys can get stuck otherwise
;; state is either 1 (ON) or 0 (OFF)
(def (grab! state)
	(using ((state :~ (in-range? 0 2)))
		(when (fxxor grabbed? state)
			(check-os-error (_ioctl kb EVIOCGRAB state)
											(ioctl 'kb 'EVIOCGRAB state))
			(set! grabbed? state))))
